with Ada.Calendar; use Ada.Calendar;
with Ada.Text_IO; use Ada.Text_IO;
with Ada.Command_Line; use Ada.Command_Line;
with Ada.Directories; use Ada.Directories;
with Ada.Directories.Information;
with Ada.Strings.Unbounded; use Ada.Strings.Unbounded;
with Ada.Strings.Hash;
with Ada.Containers; use Ada.Containers;
with Ada.Containers.Indefinite_Ordered_Sets;
with GNAT_Helper; use GNAT_Helper;
procedure gnatoptlink is
	
	Print_Usage : exception;
	Print_Version : exception;
	Invalid_Args : exception;
	
	procedure Usage is
		use Ada.Text_IO;
	begin
		Put_Line("Usage: gnatoptlink [switches] mainprog.ali [non-Ada-objects] [linker-options]");
		New_Line;
		Put_Line("  -a            Consider all files, even readonly ali files");
		Put_Line("  --help        Print this message");
		Put_Line("  --make-lgnarl Remake libgnarl.a for smart linking");
		Put_Line("  --make-lgnat  Remake libgnat.a for smart linking");
		Put_Line("  --time        Print total processing time");
		Put_Line("  -v            Verbose mode");
		Put_Line("  --version     Print version number and exit");
		Put_Line("  -xr           Reuse temporary files");
		Put_Line("  -xt           Do NOT delete temporary files");
		Put_Line("  -xS           Do NOT delete temporary assembly files");
	end Usage;

	function "+" (X : String) return Ada.Strings.Unbounded.Unbounded_String
		renames Ada.Strings.Unbounded.To_Unbounded_String;
	function "+" (X : Ada.Strings.Unbounded.Unbounded_String) return String
		renames Ada.Strings.Unbounded.To_String;

	Not_Found : exception;
	Parse_Error : exception;

	package String_Sets is new Ada.Containers.Indefinite_Ordered_Sets (String);
	use String_Sets;

	Main : Unbounded_String;
	Linker_Options : Unbounded_String;
	All_Files : Boolean := False;
	Sources : String_Sets.Set;
	Obj_Dirs : String_Sets.Set;

	procedure Read_Binder is
	begin
		-- .adb for linker options
		declare
			F : File_Type;
			Begin_Mark : constant String := "--  BEGIN Object file/option list";
			End_Mark : constant String := "--  END Object file/option list";
			Header : constant String := "   --   ";
		begin
			Open(F, In_File, "b~" & (+Main) & ".adb");
			loop
				declare
					L : constant String := Get_Line(F);
				begin
					if L'Length >= Begin_Mark'Length then
						exit when L(L'First .. L'First + Begin_Mark'Length - 1) = Begin_Mark;
					end if;
				end;
			end loop;
			loop
				declare
					Buf : String(1 .. 256);
					Last : Natural;
				begin
					Get_Line (F, Buf, Last);
					if Last >= End_Mark'Length
						and then Buf(1 .. End_Mark'Length) = End_Mark
					then
						exit;
					elsif Last >= Header'Length
						and then Buf(1 .. Header'Length) = Header
					then
						if (not All_Files or else (Buf(Header'Length + 1 .. Last) /= "-lgnat" 
							and then Buf(Header'Length + 1 .. Last) /= "-lgnarl"))
							and then not (Last > 2 and then Buf(Last - 1) = '.' and then Buf(Last) = 'o')
						then
							Append(Linker_Options, Buf(Header'Length + 1 .. Last) & ' ');
							if Buf(Header'Length + 1 .. Header'Length + 2) = "-L" then
								Include(Obj_Dirs, Buf(Header'Length + 3 .. Last));
							end if;
						end if;
					else
						raise Parse_Error;
					end if;
				end;
			end loop;
			Close(F);
		end;
		-- .ads for sources
		declare
			F : File_Type;
			Begin_Mark : constant String := "   --  BEGIN ELABORATION ORDER";
			End_Mark : constant String := "   --  END ELABORATION ORDER";
			Header : constant String := "   --  ";
		begin
			Open(F, In_File, "b~" & (+Main) & ".ads");
			loop
				declare
					L : constant String := Get_Line(F);
				begin
					if L'Length >= Begin_Mark'Length then
						exit when L(L'First .. L'First + Begin_Mark'Length - 1) = Begin_Mark;
					end if;
				end;
			end loop;
			loop
				declare
					Buf : String(1 .. 256);
					Last : Natural;
				begin
					Get_Line (F, Buf, Last);
					if Last >= End_Mark'Length
						and then Buf(1 .. End_Mark'Length) = End_Mark
					then
						exit;
					elsif Last >= Header'Length
						and then Buf(1 .. Header'Length) = Header
						and then Buf(Last - 1) = '%'
					then
						if Buf(Last) = 's' then
							Include(Sources, Buf(Header'Length + 1 .. Last - 2) & ".ads");
						else
							Exclude(Sources, Buf(Header'Length + 1 .. Last - 2) & ".ads");
							Include(Sources, Buf(Header'Length + 1 .. Last - 2) & ".adb");
						end if;
					else
						raise Parse_Error;
					end if;
				end;
			end loop;
			Close(F);
		end;
	end Read_Binder;

	function Find_ALI (Source : String) return String is
		I : String_Sets.Cursor := First(Obj_Dirs);
	begin
		while Has_Element(I) loop
			declare
				F : String := Compose(Element(I), Base_Name(Source), "ali");
			begin
				if Exists(F) then
					return F;
				end if;
			end;
			Next(I);
		end loop;
		raise Not_Found with Compose("", Base_Name(Source), "ali");
	end Find_ALI;
	
	function Read_ALI(Name : String) return String is
		F : File_Type;
		G : Unbounded_String;
	begin
		Open(F, In_File, Name);
		while not End_Of_File(F) loop
			declare
				O :  String := Get_Line(F);
			begin
				if O'Length > 2 and then O(O'First .. O'First + 1) = "A " 
					and then O(O'First + 2 .. O'Last) /= "-g"
				then
					Append(G, O(O'First + 2 .. O'Last));
					Append(G, ' ');
				end if;
			end;
		end loop;
		Close(F);
		return +G;
	end Read_ALI;
	
	Temp : constant String := ".gnatoptlink";
	Temp_File : constant String := Compose(Temp, "1");
	
	Verbose : Boolean := False;
	Not_Delete_Asm : Boolean := False;
	
	procedure Compile(Lib, Source, Option : String; Already : String_Sets.Set) is
		Asm : constant String := Compose(Temp, Base_Name(Source), "s");
		ALI : constant String := Compose(Temp, Base_Name(Source), "ali");
	begin
		if (Contains(Already, Base_Name(Source) & ".0.o") or else Exists(Asm))
			and then not Exists(ALI)
		then
			return;
		end if;
		Execute("cd " & Temp & " && gcc -S " & Option &
			"-ffunction-sections -fdata-sections " &
			"-o ..\" & Asm & " " & Source, Verbose);
		declare
			F : File_Type;
			F2 : File_Type;
			Index : Natural := 0;
			Hash : Ada.Containers.Hash_Type := Ada.Strings.Hash(Source);
			H : constant array (Ada.Containers.Hash_Type range 0 .. 15) of Character := "0123456789abcdef";
			Local_To_Global : String(1 .. 9);
			procedure Assemble is
				Asm2 : constant String := Name(F2);
				Obj2 : constant String := Compose(Temp, Base_Name(Asm2), "o");
			begin
				Close(F2);
				if not Contains(Already, Simple_Name(Obj2)) then
					Execute("as """ & Asm2 & """ -o " & Obj2, Verbose);
					Execute("ar -qcS " & Lib & " " & Obj2, Verbose);
					Ada.Directories.Delete_File(Obj2);
				end if;
				if not Not_Delete_Asm then
					Ada.Directories.Delete_File(Asm2);
				end if;
				Index := Index + 1;
			end Assemble;
		begin
			Local_To_Global(9) := H(Hash mod 16); Hash := Hash / 16;
			Local_To_Global(8) := H(Hash mod 16); Hash := Hash / 16;
			Local_To_Global(7) := H(Hash mod 16); Hash := Hash / 16;
			Local_To_Global(6) := H(Hash mod 16); Hash := Hash / 16;
			Local_To_Global(5) := H(Hash mod 16); Hash := Hash / 16;
			Local_To_Global(4) := H(Hash mod 16); Hash := Hash / 16;
			Local_To_Global(3) := H(Hash mod 16); Hash := Hash / 16;
			Local_To_Global(2) := H(Hash mod 16);
			Local_To_Global(1) := 'L';
			Open(F, In_File, Asm, "wcem=8");
			declare
				Global_Mark : constant String := ".globl ";
				Section_Mark : constant String := Character'Val(9) & ".section";
				Def_Mark : constant String := Character'Val(9) & ".def";
				File_Directive : constant String := Get_Line(F);
			begin
				while not End_Of_File(F) loop
					declare
						L : constant String := Get_Line(F);
					begin
						if L'Length = 0 then
							null;
						elsif L'Length > Section_Mark'Length and then L(Section_Mark'Range) = Section_Mark then
							if Is_Open(F2) then
								Assemble;
							end if;
							declare
								IndexS : constant String := Natural'Image(Index);
								Asm2 : constant String := Compose(Temp, Base_Name(Asm) & '.' & IndexS(2 .. IndexS'Last), "s");
								SF : Positive;
								SL : Positive;
							begin
								SF := Section_Mark'Last + 1;
								SF_Search : while L(SF - 1) /= '$' loop
									SF := SF + 1;
									if SF >= L'Length then
										SF := Section_Mark'Last + 1;
										loop
											case L(SF) is
												when Character'Val(9) | ' ' =>
													SF := SF + 1;
												when others => exit;
											end case;
										end loop;
										exit SF_Search;
									end if;
								end loop SF_Search;
								SL := SF;
								while L(SL + 1) /= ',' loop
									SL := SL + 1;
								end loop;
								Create(F2, Out_File, Asm2, "wcem=8");
								Put_Line(F2, File_Directive);
								Put_Line(F2, ".global _" & L(SF .. SL));
								Put_Line(F2, L);
							end;
						elsif L'Length > Global_Mark'Length and then L(Global_Mark'Range) = Global_Mark then
							null;
						elsif L'Length > Def_Mark'Length and then L(Def_Mark'Range) = Def_Mark then
							null;
						elsif L(1) = 'L' and then L(L'Last) = ':' then
							Put_Line(F2, ".global " & Local_To_Global & L(1 .. L'Last - 1));
							Put_Line(F2, Local_To_Global & L);
						else
							for I in L'Range loop
								if L(I) = 'L' and then (I = 1 or else (
									L(I - 1) /= '_'
									and then L(I - 1) not in '0' .. '9'
									and then L(I - 1) not in 'A' .. 'Z'
									and then L(I - 1) not in 'a' .. 'z'))
								then
									Put(F2, Local_To_Global);
									Put(F2, 'L');
								else
									Put(F2, L(I));
								end if;
							end loop;
							New_Line(F2);
						end if;
					end;
				end loop;
				if Is_Open(F2) then
					Assemble;
				end if;
			end;
			Close(F);
		end;
		Delete_File(ALI); -- finish mark
	end Compile;
	
	Version_Num : constant Version_Num_String := Extract_Version_Num(Version);
	Standard_Source_Path : constant String := "C:\mingw\lib\gcc\i686-pc-mingw32\" & Version_Num & "\adainclude";
	Standard_Library_Path : constant String := "C:\mingw\lib\gcc\i686-pc-mingw32\" & Version_Num & "\adalib";

	procedure Move(From, To, Obj : String) is
		F : constant String := Full_Name(From);
		T : constant String := Full_Name(To);
	begin
		Execute("cd " & Temp & " && ar -x " & F & " " & Obj, Verbose);
		Execute("cd " & Temp & " && ar -rcS " & T & " " & Obj, Verbose);
		Delete_File(Compose(Temp, Obj));
	end Move;

	procedure Rebuild_Library (Name : in String) is
		Lib : constant String := Simple_Name(Name);
		File : File_Type;
		Objs : String_Sets.Set;
		Already : String_Sets.Set;
		I : String_Sets.Cursor;
	begin
		-- list up
		Execute("ar -t " & Name & " >" & Temp_File, Verbose);
		Open(File, In_File, Temp_File);
		while not End_Of_File(File) loop
			Include(Objs, Get_Line(File));
		end loop;
		Close(File);
		-- reuse
		if Exists(Lib) then
			Execute("ar -t " & Lib & " >" & Temp_File, Verbose);
			Open(File, In_File, Temp_File);
			while not End_Of_File(File) loop
				Include(Already, Get_Line(File));
			end loop;
			Close(File);
		end if;
		-- processing
		I := First(Objs);
		while Has_Element(I) loop
			declare
				Obj : constant String := Element(I);
				Source : String := Compose(Standard_Source_Path, Base_Name(Obj), "adb");
			begin
				if not Exists(Source) then
					Source(Source'Last) := 's';
				end if;
				if not Exists(Source) or else Obj = "a-except.o" then
					Move(From => Name, To => Lib, Obj => Obj);
				else
					declare
						Option : String := Read_ALI(Compose(Standard_Library_Path, Base_Name(Obj), "ali"));
					begin
						Compile(Lib => Lib, Source => Source, Option => Option, Already => Already);
					end;
				end if;
			end;
			Next(I);
		end loop;
		-- index
		Execute("ranlib " & Lib, Verbose);
	end Rebuild_Library;

	libgnat : constant String := Compose(Standard_Library_Path, "libgnat.a");
	libgnarl : constant String := Compose(Standard_Library_Path, "libgnarl.a");
	Not_Delete_Temp : Boolean := False;
	Reuse_Temp : Boolean := False;
	Show_Time : Boolean := False;
	Start : Time := Clock;
	Rebuild_lgnat : Boolean := False;
	Rebuild_lgnarl : Boolean := False;
	Lib : constant String := Compose(Temp, "all.a");
begin
	if Ada.Command_Line.Argument_Count = 0 then
		raise Print_Usage;
	end if;
	for I in 1 .. Ada.Command_Line.Argument_Count loop
		declare
			Arg : constant String := Ada.Command_Line.Argument(I);
		begin
			if Main /= "" then
				Append(Linker_Options, Arg & ' ');
			elsif Arg = "-a" then
				All_Files := True;
			elsif Arg = "-v" then
				Verbose := True;
			elsif Arg = "-xr" then
				Reuse_Temp := True;
			elsif Arg = "-xt" then
				Not_Delete_Temp := True;
			elsif Arg = "-xS" then
				Not_Delete_Temp := True;
				Not_Delete_Asm := True;
			elsif Arg = "--help" then
				raise Print_Usage;
			elsif Arg = "--make-lgnarl" then
				Rebuild_lgnarl := True;
			elsif Arg = "--make-lgnat" then
				Rebuild_lgnat := True;
			elsif Arg = "--time" then
				Show_Time := True;
			elsif Arg = "--version" then
				raise Print_Version;
			elsif Arg (Arg'First) = '-' then
				raise Invalid_Args;
			else
				Main := +Arg;
			end if;
		end;
	end loop;
	-- make temp
	Create_Path(Temp);
	-- start
	if Rebuild_lgnat then
		Rebuild_Library (libgnat);
	end if;
	if Rebuild_lgnarl then
		Rebuild_Library (libgnarl);
	end if;
	if not Rebuild_lgnat and not Rebuild_lgnarl then
		if Extension(+Main) = "ali" then
			Main := +Base_Name(+Main);
		end if;
		-- binder
		Execute("gcc -c -gnatA -gnatWb -gnatiw -gnatws -O2 b~" & (+Main) & ".adb", Verbose);
		Read_Binder;
		Include(Obj_Dirs, Standard_Library_Path);
		Exclude(Sources, "ada.exceptions.adb");
		-- make section-ed objects
		if not (Reuse_Temp and then Exists(Lib)) then
			if Exists(Lib) then
				Delete_File(Lib);
			end if;
			declare
				I : String_Sets.Cursor := First(Sources);
			begin
				while Has_Element(I) loop
					declare
						E : Unbounded_String := +Element(I);
						G : Unbounded_String := Null_Unbounded_String;
					begin
						-- to 8.3
						if (Length(E) > 4 and then Slice(E, 1, 4) = "ada.")
							or else (Length(E) > 5 and then Slice(E, 1, 5) = "gnat.")
							or else (Length(E) > 11 and then Slice(E, 1, 11) = "interfaces.")
							or else (Length(E) > 7 and then Slice(E, 1, 7) = "system.")
						then
							if not All_Files then
								goto Skip;
							end if;
							Execute("gnatkr " & (+E) & " >" & Temp_File, Verbose);
							declare
								F : File_Type;
							begin
								Open(F, In_File, Temp_File);
								E := +Get_Line(F);
								Close(F);
							end;
						end if;
						-- get options from ali file
						declare
							ALI : String := Find_ALI (+E);
						begin
							-- if .ali was read-only then skip
							if not All_Files and Information.Is_Read_Only (ALI) then -- ** Ada.Directories.Information.Is_Read_Only **
								goto Skip;
							end if;
							Append(G, Read_ALI(ALI));
						end;
						-- recompile
						declare
							S : Unbounded_String;
						begin
							if Exists(+E) then
								S := +Compose("..", +E);
							elsif Exists(Compose(Standard_Source_Path, +E)) then
								S := +Compose(Standard_Source_Path, +E);
							else
								S := E;
							end if;
							Compile(Lib => Lib, Source => +S, Option => +G, Already => String_Sets.Empty_Set);
						end;
					end;
					<<Skip>>
					Next(I);
				end loop;
			end;
		end if;
		--  move from libgnat.a
		if All_Files then
			declare
				F : File_Type;
			begin
				Move(From => libgnat, To => lib, Obj => "a-except.o");
				Execute("ar -t " & libgnat & " >" & Temp_File, Verbose);
				Open(F, In_File, Temp_File);
				while not End_Of_File(F) loop
					declare
						Obj : constant String := Get_Line(F);
					begin
						if Obj (2) /= '-' 
							and then not Exists(Compose(Standard_Source_Path, Base_Name(Obj), "ads"))
							and then not Exists(Compose(Standard_Source_Path, Base_Name(Obj), "adb"))
						then
							Move(From => libgnat, To => lib, Obj => Obj);
						end if;
					end;
				end loop;
				Close(F);
			end;
		end if;
		Execute("ranlib " & Lib, Verbose);
		-- link
		Execute("gcc b~" & (+Main) & ".o -o " & (+Main) & ".exe " & Lib & ' ' & (+Linker_Options), Verbose);
	end if;
	-- end
	if Show_Time then
		Put_Line("Time:" & Duration'Image(Clock - Start) & " sec");
	end if;
	-- remove temp
	if not Not_Delete_Temp then
		Delete_Tree(Temp);
	end if;
exception
	when Print_Usage => Usage;
	when Print_Version => Ada.Text_IO.Put_Line(Version);
	when Invalid_Args => Ada.Text_IO.Put_Line(Ada.Text_IO.Current_Error, "Invalid command line argument.");
end gnatoptlink;
